/* Copyright (C) 2003 Timo Sirainen */

#include "lib.h"
#include "buffer.h"
#include "str.h"
#include "oracle-storage.h"
#include "oracle-mail.h"

struct mail_fetch_context {
	struct oracle_mailbox *obox;
	sqlo_stmt_handle_t sth;

	struct oracle_mail mail;
	int all_found, next, failed;
};

static void str_build_fetch_fields(string_t *str, enum mail_fetch_field fields)
{
	int cache;

	if (fields & MAIL_FETCH_FLAGS)
		str_append(str, ", answered, flagged, deleted, seen, draft");
	if (fields & MAIL_FETCH_RECEIVED_DATE)
		str_append(str, ", TO_CHAR(internal_date, 'yyyymmddhh24miss')");
	if (fields & MAIL_FETCH_SIZE)
		str_append(str, ", message_size");

	if (fields & (MAIL_FETCH_STREAM_HEADER|MAIL_FETCH_STREAM_BODY))
		str_append(str, ", data");

	cache = (fields & (MAIL_FETCH_MESSAGE_PARTS |
			   MAIL_FETCH_DATE |
			   MAIL_FETCH_IMAP_BODY |
			   MAIL_FETCH_IMAP_BODYSTRUCTURE |
			   MAIL_FETCH_IMAP_ENVELOPE)) != 0;
	if (cache) {
		if (fields & MAIL_FETCH_MESSAGE_PARTS)
			str_append(str, ", messageparts");
		if (fields & MAIL_FETCH_DATE) {
			str_append(str,
				   ", TO_CHAR(sent_date, 'yyyymmddhh24miss')");
		}
		if (fields & MAIL_FETCH_IMAP_BODY)
			str_append(str, ", imap_body");
		if (fields & MAIL_FETCH_IMAP_BODYSTRUCTURE)
			str_append(str, ", imap_bodystructure");
		if (fields & MAIL_FETCH_IMAP_ENVELOPE)
			str_append(str, ", imap_envelope");
	}

	str_append(str, " FROM mails");
	if (cache)
		str_append(str, ", mail_cache");
	if (fields & (MAIL_FETCH_STREAM_HEADER|MAIL_FETCH_STREAM_BODY))
		str_append(str, ", mail_body");

	str_append(str, " WHERE ");
	if (cache) {
		str_append(str, "mails.mail_body_id = "
			   "mail_cache.mail_body_id (+) AND ");
	}
	if (fields & (MAIL_FETCH_STREAM_HEADER|MAIL_FETCH_STREAM_BODY))
		str_append(str, "mails.mail_body_id = mail_body.id AND ");
}

static int str_build_define_output(struct mail_fetch_context *ctx,
				   enum mail_fetch_field fields)
{
	struct oracle_storage *ostorage = ctx->obox->ostorage;
	int pos = 1;

	if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_UIN,
			      &ctx->mail.mail.uid, sizeof(unsigned int)) < 0)
		return FALSE;

	if (fields & MAIL_FETCH_FLAGS) {
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_AFC,
				      &ctx->mail.answered, 1) < 0)
			return FALSE;
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_AFC,
				      &ctx->mail.flagged, 1) < 0)
			return FALSE;
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_AFC,
				      &ctx->mail.deleted, 1) < 0)
			return FALSE;
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_AFC,
				      &ctx->mail.seen, 1) < 0)
			return FALSE;
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_AFC,
				      &ctx->mail.draft, 1) < 0)
			return FALSE;
	}

	if (fields & MAIL_FETCH_RECEIVED_DATE) {
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_STR,
				      &ctx->mail.received_date,
				      sizeof(ctx->mail.received_date)) < 0)
			return FALSE;
	}
	if (fields & MAIL_FETCH_SIZE) {
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_UIN,
				      &ctx->mail.message_size,
				      sizeof(ctx->mail.message_size)) < 0)
			return FALSE;
	}

	if (fields & (MAIL_FETCH_STREAM_HEADER|MAIL_FETCH_STREAM_BODY)) {
		if (orasql_define_pos(ostorage, ctx->sth, pos++, SQLOT_BLOB,
				      &ctx->mail.lob, 0) < 0)
			return FALSE;
	}

	return TRUE;
}

struct mail_fetch_context *
oracle_storage_fetch_init(struct mailbox *box,
			  enum mail_fetch_field wanted_fields, int *update_seen,
			  const char *messageset, int uidset)
{
	struct oracle_mailbox *obox = (struct oracle_mailbox *)box;
        struct mail_fetch_context *ctx;
	sqlo_stmt_handle_t sth;
	string_t *str;
	int ret;

	str = t_str_new(512);
	str_append(str, "SELECT imap_uid");
	str_build_fetch_fields(str, wanted_fields);

	ret = oracle_messageset_build(obox, str, messageset, uidset, TRUE);
	if (ret < 0)
		return NULL;

	sth = sqlo_prepare(obox->ostorage->dbh, str_c(str));
	if (sth < 0) {
		orasql_set_error(obox->ostorage, "sqlo_prepare()");
		return NULL;
	}

	ctx = i_new(struct mail_fetch_context, 1);
	ctx->obox = obox;
	ctx->sth = sth;
	ctx->all_found = ret > 0;

	oracle_mail_init(obox, &ctx->mail);

	if (!str_build_define_output(ctx, wanted_fields) ||
	    orasql_execute(obox->ostorage, ctx->sth) < 0) {
		i_free(ctx);
		(void)orasql_close(obox->ostorage, sth);
		return NULL;
	}

	return ctx;
}

int oracle_storage_fetch_deinit(struct mail_fetch_context *ctx, int *all_found)
{
	int failed = ctx->failed;

	*all_found = ctx->all_found;
	i_free(ctx);
	return !failed;
}

struct mail *oracle_storage_fetch_next(struct mail_fetch_context *ctx)
{
	if (!ctx->next)
		ctx->next = TRUE;
	else {
		switch (orasql_fetch(ctx->obox->ostorage, ctx->sth)) {
		case -1:
			ctx->failed = TRUE;
			return NULL;
		case 0:
			return NULL;
		}
	}

	if (oracle_mail_next(&ctx->mail) < 0) {
		ctx->failed = TRUE;
		return NULL;
	}

	return &ctx->mail.mail;
}

struct mail *oracle_storage_fetch_uid(struct mailbox *box, unsigned int uid,
				      enum mail_fetch_field wanted_fields)
{
	struct oracle_mailbox *obox = (struct oracle_mailbox *)box;
	sqlo_stmt_handle_t sth;
	string_t *str;

	str = t_str_new(256);
	str_append(str, "SELECT imap_uid");
	str_build_fetch_fields(str, wanted_fields);
	str_printfa(str, "imap_uid = %u", uid);

	sth = sqlo_prepare(obox->ostorage->dbh, str_c(str));
	if (sth < 0) {
		orasql_set_error(obox->ostorage, "sqlo_prepare()");
		return NULL;
	}

	/*if (!str_build_define_output(ctx, wanted_fields))
		return NULL;

	if (!orasql_mailbox_fetch_one(obox, sth))
		return NULL;*/

	return NULL;
}

struct mail *oracle_storage_fetch_seq(struct mailbox *box, unsigned int seq,
				      enum mail_fetch_field wanted_fields)
{
	struct oracle_mailbox *obox = (struct oracle_mailbox *)box;
	const unsigned char *uid_arr;
	unsigned int uid;

	uid_arr = buffer_get_data(obox->uids, NULL);
	uid = seq > 0 && seq <= obox->synced_messages ? uid_arr[seq-1] : 0;
	return uid == 0 ? NULL :
		oracle_storage_fetch_uid(box, uid, wanted_fields);
}
