/* Copyright (C) 2002-2003 Timo Sirainen */

#include "lib.h"
#include "buffer.h"
#include "str.h"
#include "oracle-storage.h"

struct messageset_context {
	struct oracle_mailbox *obox;

	unsigned int messages_count, last_seq;
	unsigned int num1, num2, seq;

	const unsigned int *uid_arr;
	const char *messageset, *p;
	int uidset, skip_expunged;

	int expunges_found, ret;
	const char *error;
};

static unsigned int get_next_number(const char **str)
{
	unsigned int num;

	num = 0;
	while (**str != '\0') {
		if (**str < '0' || **str > '9')
			break;

		num = num*10 + (**str - '0');
		(*str)++;
	}

	return num;
}

static int messageset_parse_next(struct messageset_context *ctx)
{
	unsigned int num;

	if (ctx->p == NULL) {
		/* num1..num2 already set.  */
		ctx->p = "";
		return TRUE;
	}

	if (*ctx->p == '*') {
		/* last message */
		ctx->num1 = (unsigned int)-1;
		ctx->p++;
	} else {
		ctx->num1 = get_next_number(&ctx->p);
		if (ctx->num1 == 0) {
			ctx->error = t_strconcat("Invalid messageset: ",
						 ctx->messageset, NULL);
			return FALSE;
		}
	}

	if (*ctx->p != ':')
		ctx->num2 = ctx->num1;
	else {
		/* first:last range */
		ctx->p++;

		if (*ctx->p == '*') {
			ctx->num2 = (unsigned int)-1;
			ctx->p++;
		} else {
			ctx->num2 = get_next_number(&ctx->p);
			if (ctx->num2 == 0) {
				ctx->error = t_strconcat("Invalid messageset: ",
							 ctx->messageset, NULL);
				return FALSE;
			}
		}
	}

	if (*ctx->p == ',')
		ctx->p++;
	else if (*ctx->p != '\0') {
		ctx->error = t_strdup_printf("Unexpected char '%c' "
					     "with messageset: %s",
					     *ctx->p, ctx->messageset);
		return FALSE;
	}

	if (ctx->num1 > ctx->num2) {
		/* swap, as specified by RFC-3501 */
		unsigned int temp = ctx->num1;
		ctx->num1 = ctx->num2;
		ctx->num2 = temp;
	}

	num = ctx->num2 == (unsigned int)-1 ? ctx->num1 : ctx->num2;
	if (num > ctx->messages_count && !ctx->uidset &&
	    num != (unsigned int)-1) {
		ctx->error = t_strdup_printf("Message sequence %u "
					     "larger than message count (%u)",
					     num, ctx->messages_count);
		return FALSE;
	}

	return TRUE;
}

static unsigned int get_last_seq(struct messageset_context *ctx)
{
	unsigned int i;

	if (ctx->last_seq == 0) {
		for (i = ctx->messages_count; i > 0; i++) {
			if (ctx->uid_arr[i-1] != 0) {
				ctx->last_seq = i;
				break;
			}
		}
	}

	return ctx->last_seq;
}

static void uidset_init(struct messageset_context *ctx)
{
	if (ctx->num1 == (unsigned int)-1)
		ctx->num1 = get_last_seq(ctx);

	if (ctx->num2 == (unsigned int)-1)
		ctx->num2 = ctx->obox->next_uid-1;
}

static void seqset_init(struct messageset_context *ctx)
{
	unsigned int seq1, seq2;

	seq1 = ctx->num1 == (unsigned int)-1 ? ctx->messages_count : ctx->num1;
	seq2 = ctx->num2 == (unsigned int)-1 ? ctx->messages_count : ctx->num2;

	ctx->num1 = ctx->uid_arr[seq1-1];
	while (ctx->num1 == 0 && seq1 < seq2)
		ctx->num1 = ctx->uid_arr[seq1++];

	ctx->num2 = ctx->uid_arr[seq2-1];
	while (ctx->num2 == 0 && seq2 > seq1)
		ctx->num2 = ctx->uid_arr[--seq2 - 1];
}

static int is_expunges(struct messageset_context *ctx)
{
	// FIXME
	/*unsigned int i;

	for (i = ctx->num1-1; i < ctx->num2; i++) {
		if (ctx->uid_arr[i] == 0)
			return TRUE;
	}*/
	return FALSE;
}

int oracle_messageset_build(struct oracle_mailbox *obox, string_t *str,
			    const char *messageset, int uidset,
			    int skip_expunged)
{
	struct messageset_context ctx;
	int first = TRUE;

	ctx.obox = obox;
	ctx.messages_count = obox->synced_messages;
	ctx.uid_arr = buffer_get_data(obox->uids, NULL);
	ctx.p = ctx.messageset = messageset;
	ctx.uidset = uidset;

	str_append_c(str, '(');
	for (;;) {
		if (ctx.p != NULL && *ctx.p == '\0') {
			/* finished */
			ctx.ret = 1;
			break;
		}

		if (!messageset_parse_next(&ctx)) {
			mail_storage_set_syntax_error(obox->storage,
						      "%s", ctx.error);
			return -2;
		}

		if (uidset)
			uidset_init(&ctx);
		else
			seqset_init(&ctx);

		if (!skip_expunged && (ctx.num1 == 0 || is_expunges(&ctx))) {
			/* we wish to abort if there's any
			   expunged messages */
			return 0;
		}

		if (!first)
			str_append(str, " OR");
		else
			first = FALSE;

		if (ctx.num1 == 0 || ctx.num2 == 0) {
			/* range doesn't exist */
		} else if (ctx.num1 == ctx.num2)
			str_printfa(str, "imap_uid = %u", ctx.num1);
		else {
			str_printfa(str, "(imap_uid >= %u AND imap_uid <= %u)",
				    ctx.num1, ctx.num2);
		}
	}
	str_append_c(str, ')');

	return 1;
}
