/* Copyright (C) 2003 Timo Sirainen */

#include "lib.h"
#include "istream.h"
#include "ostream-lob.h"
#include "oracle-storage.h"
#include "mail-save.h"

#include <time.h>

struct mail_save_context {
	struct oracle_mailbox *obox;
	int transaction, failed;
};

struct mail_save_context *
oracle_storage_save_init(struct mailbox *box, int transaction)
{
	struct oracle_mailbox *obox = (struct oracle_mailbox *)box;
        struct mail_save_context *ctx;

	ctx = i_new(struct mail_save_context, 1);
	ctx->obox = obox;
	ctx->transaction = transaction;
	return ctx;
}

int oracle_storage_save_deinit(struct mail_save_context *ctx, int rollback)
{
	int ret = TRUE;

	if (ctx->failed)
		ret = orasql_rollback(ctx->obox->ostorage);
	else if (ctx->transaction) {
		if (rollback)
			ret = orasql_rollback(ctx->obox->ostorage) >= 0;
		else
			ret = orasql_commit(ctx->obox->ostorage) >= 0;
	}
	i_free(ctx);
	return ret;
}

int oracle_storage_save_next(struct mail_save_context *ctx,
			     const struct mail_full_flags *flags,
			     time_t received_date,
			     int timezone_offset __attr_unused__,
			     struct istream *data)
{
	struct oracle_storage *ostorage = ctx->obox->ostorage;
	struct ostream *output;
	sqlo_stmt_handle_t sth;
	sqlo_lob_desc_t lob;
	unsigned int id;
	struct tm *tm;
	const char *str;
	int ret;

	/* read mail body */
	if (sqlo_alloc_lob_desc(ostorage->dbh, &lob) < 0) {
		orasql_set_error(ostorage, "sqlo_alloc_lob_desc()");
		ctx->failed = TRUE;
		return FALSE;
	}

	sth = orasql_prepare(ostorage,
			     "INSERT INTO mail_body (data) "
			     "VALUES (EMPTY_BLOB()) "
			     "RETURNING id, data INTO :id, %p:blob", &lob);
	if (sqlo_bind_by_name(sth, ":id", SQLOT_UIN, &id, sizeof(id),
			      NULL, 0) != SQLO_SUCCESS) {
		mail_storage_set_critical(&ostorage->storage,
			"sqlo_bind_by_name(mail_body, :id) failed: %s",
			sqlo_geterror(ostorage->dbh));
		sth = -1;
	}

	if ((ret = orasql_execute(ostorage, sth)) <= 0) {
		if (ret == 0) {
			mail_storage_set_critical(&ostorage->storage,
				"APPEND: BLOB not returned?");
		}
		ctx->failed = TRUE;
	} else {
		output = o_stream_create_lob(ostorage->dbh, lob,
					     data_stack_pool);
		ctx->failed = !mail_storage_save(&ostorage->storage,
						 ctx->obox->box.name,
						 data, output,
						 TRUE, NULL, NULL);
		o_stream_unref(output);
	}

	sqlo_free_lob_desc(ostorage->dbh, &lob);
	if (orasql_close(ostorage, sth) < 0)
		ctx->failed = TRUE;

	if (ctx->failed)
		return FALSE;

	/* save mail metadata */
	tm = localtime(&received_date);
	str = t_strdup_printf("%04d%02d%02d%02d%02d%02d",
			      tm->tm_year, tm->tm_mon+1, tm->tm_mday,
			      tm->tm_hour, tm->tm_min, tm->tm_sec);

	if (orasql_exec(ostorage, "INSERT INTO mails "
			"(mailbox_id, mail_body_id, internal_date) "
			"VALUES (%u:mailbox, %u:body, "
			" TO_DATE(%s:intdate, 'yyyymmddhh24miss'))",
			ctx->obox->id, id, str) < 0) {
		ctx->failed = TRUE;
		return FALSE;
	}

	if (!ctx->transaction) {
		if (!orasql_commit(ostorage)) {
			ctx->failed = TRUE;
			return FALSE;
		}
	}
	return TRUE;
}
