/* Copyright (C) 2003 Timo Sirainen */

#include "lib.h"
#include "oracle-storage.h"

void orasql_set_error(struct oracle_storage *storage, const char *error)
{
	mail_storage_set_critical(&storage->storage, "%s failed: %s",
				  error, sqlo_geterror(storage->dbh));
}

static const char *param_get(const char *str)
{
	const char *p = str;

	while (*p != '\0' && *p != ' ' && *p != ',')
		p++;

	return *p == '\0' ? str : t_strdup_until(str, p);
}

static sqlo_stmt_handle_t
orasql_preparev(struct oracle_storage *storage, const char *fmt, va_list args)
{
	const char *p, *param;
	char *sql, *sqlp;
	sqlo_stmt_handle_t sth;
	int failed = FALSE;

	/* first we need the sql */
	sql = sqlp = t_malloc(strlen(fmt)+1);
	for (p = fmt; *p != '\0'; p++) {
		if (*p != '%' || p[1] == '\0' || p[2] != ':')
			*sqlp++ = *p;
		else
			p++;
	}
	*sqlp = '\0';

	sth = sqlo_prepare(storage->dbh, sql);
	if (sth < 0) {
		mail_storage_set_critical(&storage->storage,
					  "sqlo_prepare(%s) failed: %s",
					  sql, sqlo_geterror(storage->dbh));
		return -1;
	}

	/* assign the variables */
	for (p = fmt; *p != '\0' && !failed; p++) {
		if (*p != '%' || p[1] == '\0' || p[2] != ':')
			continue;

		p++;
		param = param_get(p+1);
		switch (*p) {
		case 'u': {
			unsigned int arg = va_arg(args, unsigned int);
			int *i;

			i = t_new(int, 1); *i = arg;
			if (sqlo_bind_by_name(sth, param, SQLOT_UIN,
					      i, sizeof(unsigned int),
					      NULL, 0) == SQLO_SUCCESS)
				break;

			mail_storage_set_critical(&storage->storage,
				"sqlo_bind_by_name(%s, %s, %u) failed: %s",
				sql, param, arg, sqlo_geterror(storage->dbh));
			failed = TRUE;
			break;
		}
		case 's': {
			const char *arg = va_arg(args, const char *);

			if (sqlo_bind_by_name(sth, param, SQLOT_STR,
					      arg, strlen(arg)+1,
					      NULL, 0) == SQLO_SUCCESS)
				break;

			mail_storage_set_critical(&storage->storage,
				"sqlo_bind_by_name(%s, %s, %s) failed: %s",
				sql, param, arg, sqlo_geterror(storage->dbh));
			failed = TRUE;
			break;
		}
		case 'p': {
			sqlo_lob_desc_t *arg = va_arg(args, sqlo_lob_desc_t *);

			if (sqlo_bind_by_name(sth, param, SQLOT_BLOB,
					      arg, 0, NULL, 0) == SQLO_SUCCESS)
				break;

			mail_storage_set_critical(&storage->storage,
				"sqlo_bind_by_name(%s, %s) failed: %s",
				sql, param, sqlo_geterror(storage->dbh));
			failed = TRUE;
			break;
		}
		default:
			i_panic("Unknown SQL bind type '%c'", *p);
		}
	}

	if (failed) {
		(void)sqlo_close(sth);
		sth = -1;
	}

	return sth;
}

sqlo_stmt_handle_t
orasql_prepare(struct oracle_storage *storage, const char *fmt, ...)
{
	va_list args;
        sqlo_stmt_handle_t ret;

	va_start(args, fmt);
	ret = orasql_preparev(storage, fmt, args);
	va_end(args);
	return ret;
}

sqlo_stmt_handle_t
orasql_define_output(struct oracle_storage *storage, sqlo_stmt_handle_t sth,
		     const char *fmt, ...)
{
	va_list args;
	const char *p;
	int pos = 1, failed = FALSE;

	if (sth < 0)
		return -1;

	va_start(args, fmt);
	for (p = fmt; *p != '\0' && !failed; p++) {
		if (*p != '%')
			i_panic("orasql_define_output(%s): Invalid fmt", fmt);
		switch (*++p) {
		case 'u': {
			unsigned int *arg = va_arg(args, unsigned int *);

			if (sqlo_define_by_pos(sth, pos, SQLOT_UIN,
					       arg, sizeof(unsigned int),
					       NULL, NULL, 0) == SQLO_SUCCESS)
				break;

			mail_storage_set_critical(&storage->storage,
						  "sqlo_define_by_pos(%d) "
						  "failed: %s", pos,
						  sqlo_geterror(storage->dbh));
			failed = TRUE;
			break;
		}
		default:
			i_panic("Unknown SQL define type '%c'", *p);
		}

		pos++;
	}

	va_end(args);

	if (failed) {
		(void)sqlo_close(sth);
		sth = -1;
	}

	return sth;
}

int orasql_define_pos(struct oracle_storage *storage, sqlo_stmt_handle_t sth,
		      int pos, int type, void *arg, unsigned int arg_size)
{
	if (sqlo_define_by_pos(sth, pos, type, arg, arg_size,
			       NULL, NULL, 0) == SQLO_SUCCESS)
		return 0;

	mail_storage_set_critical(&storage->storage,
				  "sqlo_define_by_pos(%d,%d) "
				  "failed: %s", pos, type,
				  sqlo_geterror(storage->dbh));
	return -1;
}

int orasql_execute(struct oracle_storage *storage, sqlo_stmt_handle_t sth)
{
	if (sth < 0)
		return -1;

	switch (sqlo_execute(sth, 1)) {
	case SQLO_SUCCESS:
		return 1;
	case SQLO_NO_DATA:
		return 0;
	default:
		orasql_set_error(storage, "sqlo_execute()");
		(void)sqlo_close(sth);
		return -1;
	}
}

int orasql_fetch(struct oracle_storage *storage, sqlo_stmt_handle_t sth)
{
	if (sth < 0)
		return -1;

	switch (sqlo_fetch(sth, 1)) {
	case SQLO_SUCCESS:
		return 1;
	case SQLO_NO_DATA:
		return 0;
	default:
		orasql_set_error(storage, "sqlo_fetch()");
		(void)sqlo_close(sth);
		return -1;
	}
}

int orasql_fetch_one(struct oracle_storage *storage, sqlo_stmt_handle_t sth)
{
	int ret;

	if (sth < 0)
		return -1;

	ret = orasql_execute(storage, sth);
	if (ret <= 0)
		return ret;

	if (sqlo_close(sth) < 0) {
		orasql_set_error(storage, "sqlo_close()");
		return -1;
	}
	return ret;
}

int orasql_exec(struct oracle_storage *storage, const char *fmt, ...)
{
	va_list args;
	sqlo_stmt_handle_t sth;
	int ret;

	va_start(args, fmt);
	sth = orasql_preparev(storage, fmt, args);
	va_end(args);

	if (orasql_execute(storage, sth) < 0)
		return -1;

        ret = sqlo_prows(sth);

	if (sqlo_close(sth) < 0) {
		orasql_set_error(storage, "sqlo_close()");
		return -1;
	}
	return ret;
}

int orasql_close(struct oracle_storage *storage, sqlo_stmt_handle_t sth)
{
	if (sqlo_close(sth) < 0) {
		orasql_set_error(storage, "sqlo_close()");
		return -1;
	}
	return 0;
}

int orasql_mailbox_fetch_one(struct oracle_mailbox *obox,
			     sqlo_stmt_handle_t sth)
{
	int ret;

	if ((ret = orasql_fetch_one(obox->ostorage, sth)) <= 0) {
		if (ret == 0) {
			mail_storage_set_error(obox->storage,
					       "Mailbox doesn't exist anymore");
		}
		return FALSE;
	}

	return TRUE;
}

int orasql_rollback(struct oracle_storage *storage)
{
	if (sqlo_rollback(storage->dbh) < 0) {
		orasql_set_error(storage, "sqlo_rollback()");
		return -1;
	}
	return 0;
}

int orasql_commit(struct oracle_storage *storage)
{
	if (sqlo_commit(storage->dbh) < 0) {
		orasql_set_error(storage, "sqlo_commit()");
		return -1;
	}
	return 0;
}
