/* Copyright (C) 2003 Timo Sirainen */

#include "lib.h"
#include "buffer.h"
#include "oracle-storage.h"

#define PERFETCH_ROWS 100

static int sqlora_initialized = FALSE;

static void oracle_free(struct mail_storage *storage);

static struct mail_storage *oracle_create(const char *data, const char *user)
{
	struct oracle_storage *storage;
	sqlo_db_handle_t dbh;
	sqlo_stmt_handle_t sth;
	int ret;

	if (!sqlora_initialized) {
		if (sqlo_init(SQLO_OFF, 1, PERFETCH_ROWS) != SQLO_SUCCESS) {
			i_error("libsqlora8 initialization failed");
			return NULL;
		}
		sqlora_initialized = TRUE;
	}

	if (sqlo_connect(&dbh, data) != SQLO_SUCCESS) {
		i_error("Can't connect to oracle as %s", data);
		return NULL;
	}

	storage = i_new(struct oracle_storage, 1);
	storage->storage = oracle_storage;
	storage->dbh = dbh;

	sth = orasql_prepare(storage, "SELECT id FROM users "
			     "WHERE name = %s:name", user);
	sth = orasql_define_output(storage, sth, "%u", &storage->user_id);
	if ((ret = orasql_fetch_one(storage, sth)) <= 0) {
		if (ret == 0)
			i_error("User '%s' not in database", user);
                oracle_free(&storage->storage);
		return NULL;
	}

	storage->storage.dir = i_strdup(data);
	storage->storage.user = i_strdup(user);
	storage->storage.callbacks = i_new(struct mail_storage_callbacks, 1);
	return &storage->storage;
}

static void oracle_free(struct mail_storage *storage)
{
	struct oracle_storage *ostorage = (struct oracle_storage *)storage;

	sqlo_finish(ostorage->dbh);

	i_free(storage->dir);
	i_free(storage->user);
	i_free(storage->error);
	i_free(storage->callbacks);
	i_free(storage);
}

static int oracle_autodetect(const char *data __attr_unused__)
{
	return FALSE;
}

static void oracle_set_callbacks(struct mail_storage *storage,
				 struct mail_storage_callbacks *callbacks,
				 void *context)
{
	*storage->callbacks = *callbacks;
	storage->callback_context = context;
}

static int oracle_is_valid_create_name(struct mail_storage *storage,
				       const char *name)
{
	if (name[0] == '\0' || name[0] == storage->hierarchy_sep ||
	    name[strlen(name)-1] == storage->hierarchy_sep)
		return FALSE;

	return TRUE;
}

static const char *inbox_fix_case(struct mail_storage *storage,
				  const char *name)
{
        if (strncasecmp(name, "INBOX", 5) == 0 &&
	    (name[5] == '\0' || name[5] == storage->hierarchy_sep)) {
		/* use same case with all INBOX folders or we'll get
		   into trouble */
		name = t_strconcat("INBOX", name+5, NULL);
	}

	return name;
}

static int mailbox_traverse(struct oracle_storage *storage, const char *name,
			    unsigned int *last_id, unsigned int *level)
{
	sqlo_stmt_handle_t sth;
	const char *const *split;
	char sep[2];
	unsigned int parent_id = 0;
	int ret;

	name = inbox_fix_case(&storage->storage, name);

	*last_id = 0;
	if (level != NULL) *level = 0;

	sep[0] = storage->storage.hierarchy_sep; sep[1] = '\0';

	for (split = t_strsplit(name, sep); *split != NULL; split++) {
		sth = orasql_prepare(storage,
			parent_id == 0 ?
			"SELECT id FROM mailboxes "
			"WHERE user_id = %u:userid AND name = %s:name AND "
			"      parent_id IS NULL" :
			"SELECT id FROM mailboxes "
			"WHERE user_id = %u:userid AND name = %s:name AND "
			"      parent_id = %u:parent",
			storage->user_id, *split, parent_id);
		sth = orasql_define_output(storage, sth, "%u", last_id);
		if ((ret = orasql_fetch_one(storage, sth)) <= 0)
			return ret;

		parent_id = *last_id;
		if (level != NULL) *level += 1;
	}

	return 1;
}

static struct mailbox *
oracle_open_mailbox(struct mail_storage *storage,
		    const char *name, int readonly, int fast __attr_unused__)
{
        struct oracle_storage *ostorage = (struct oracle_storage *)storage;
	struct oracle_mailbox *obox;
	unsigned int id;
	int ret;

	ret = mailbox_traverse(ostorage, name, &id, NULL);
	if (ret < 0)
		return NULL;
	if (ret == 0) {
		mail_storage_set_error(storage, "Mailbox doesn't exist");
		return NULL;
	}

	obox = i_new(struct oracle_mailbox, 1);
	obox->box = oracle_mailbox;
	obox->storage = storage;
	obox->ostorage = ostorage;
	obox->id = id;
	obox->box.readonly = readonly;

        oracle_storage_sync(&obox->box, FALSE);
	return &obox->box;
}

static int oracle_create_mailbox(struct mail_storage *storage,
				 const char *name, int only_hierarchy)
{
        struct oracle_storage *ostorage = (struct oracle_storage *)storage;
	unsigned int id, level;
	int ret;

	mail_storage_clear_error(storage);

	name = inbox_fix_case(storage, name);
	if (!oracle_is_valid_create_name(storage, name)) {
		mail_storage_set_error(storage, "Invalid mailbox name");
		return FALSE;
	}

	if (only_hierarchy) {
		/* no need to do anything */
		return TRUE;
	}

	ret = mailbox_traverse(ostorage, name, &id, &level);
	if (ret < 0)
		return FALSE;
	if (ret > 0) {
		mail_storage_set_error(storage, "Mailbox already exists");
		return FALSE;
	}

	// FIXME: create the missing

	return FALSE;
}

static int oracle_delete_mailbox(struct mail_storage *storage,
				 const char *name)
{
        struct oracle_storage *ostorage = (struct oracle_storage *)storage;
	unsigned int id;
	int ret;

	mail_storage_clear_error(storage);

	name = inbox_fix_case(storage, name);
	if (strcasecmp(name, "INBOX") == 0) {
		mail_storage_set_error(storage, "INBOX can't be deleted.");
		return FALSE;
	}

	ret = mailbox_traverse(ostorage, name, &id, NULL);
	if (ret > 0) {
		ret = orasql_exec(ostorage, "DELETE FROM mailboxes "
				  "WHERE id = %u:id", id);
	}

	if (ret < 0)
		return FALSE;
	if (ret == 0) {
		mail_storage_set_error(storage, "Mailbox doesn't exist");
		return FALSE;
	}

	return TRUE;
}

static int oracle_rename_mailbox(struct mail_storage *storage,
				 const char *oldname, const char *newname)
{
        struct oracle_storage *ostorage = (struct oracle_storage *)storage;
	unsigned int old_id, new_id, new_level;
	int ret;

	mail_storage_clear_error(storage);

	oldname = inbox_fix_case(storage, oldname);
	if (!oracle_is_valid_create_name(storage, newname)) {
		mail_storage_set_error(storage, "Invalid mailbox name");
		return FALSE;
	}

	ret = mailbox_traverse(ostorage, oldname, &old_id, NULL);
	if (ret < 0)
		return FALSE;
	if (ret == 0) {
		mail_storage_set_error(storage, "Mailbox doesn't exist");
		return FALSE;
	}

	ret = mailbox_traverse(ostorage, newname, &new_id, &new_level);
	if (ret < 0)
		return FALSE;
	if (ret > 0) {
		mail_storage_set_error(storage, "Mailbox already exists");
		return FALSE;
	}

	// FIXME: create newname's parent if needed, update name + parent_id
	return FALSE;
}

static int oracle_get_mailbox_name_status(struct mail_storage *storage,
					  const char *name,
					  enum mailbox_name_status *status)
{
        struct oracle_storage *ostorage = (struct oracle_storage *)storage;
	unsigned int id;
	int ret;

	mail_storage_clear_error(storage);

	name = inbox_fix_case(storage, name);

	ret = mailbox_traverse(ostorage, name, &id, NULL);
	if (ret < 0)
		return FALSE;
	if (ret > 0) {
		*status = MAILBOX_NAME_EXISTS;
		return TRUE;
	} else if (!oracle_is_valid_create_name(storage, name)) {
		*status = MAILBOX_NAME_INVALID;
		return TRUE;
	} else {
		*status = MAILBOX_NAME_VALID;
		return TRUE;
	}
}

struct mail_storage oracle_storage = {
	"oracle", /* name */

	'/', /* hierarchy_sep */

	oracle_create,
	oracle_free,
	oracle_autodetect,
	oracle_set_callbacks,
	oracle_open_mailbox,
	oracle_create_mailbox,
	oracle_delete_mailbox,
	oracle_rename_mailbox,
	oracle_list_mailbox_init,
	oracle_list_mailbox_deinit,
	oracle_list_mailbox_next,
	oracle_set_subscribed,
	oracle_get_mailbox_name_status,
	mail_storage_get_last_error,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL, NULL, NULL,

	0
};

struct mailbox oracle_mailbox = {
	NULL, /* name */
	NULL, /* storage */

	oracle_storage_close,
	oracle_storage_get_status,
	oracle_storage_sync,
	oracle_storage_auto_sync,
	oracle_storage_expunge,
	oracle_storage_update_flags,
	oracle_storage_copy,
	oracle_storage_fetch_init,
	oracle_storage_fetch_deinit,
	oracle_storage_fetch_next,
	oracle_storage_fetch_uid,
	oracle_storage_fetch_seq,
        oracle_storage_search_get_sorting,
	oracle_storage_search_init,
	oracle_storage_search_deinit,
	oracle_storage_search_next,
	oracle_storage_save_init,
	oracle_storage_save_deinit,
	oracle_storage_save_next,
	mail_storage_is_inconsistency_error,

	FALSE,
	FALSE,
	FALSE
};
