/* Copyright (C) 2003 Timo Sirainen */

#include "lib.h"
#include "buffer.h"
#include "oracle-storage.h"

static int oracle_storage_sync_and_expunge(struct oracle_mailbox *obox,
					   sqlo_stmt_handle_t sth,
					   unsigned int messages)
{
	struct mail_storage_callbacks *sc;
	void *sc_context;
	const unsigned int *old_uid, *old_last_uid, *new_uid, *new_last_uid;
	unsigned int uid, seq;
	buffer_t *new_uids;
	size_t size;
	int ret;

	sc = obox->storage->callbacks;
	sc_context = obox->storage->callback_context;

	sth = orasql_define_output(obox->ostorage, sth, "%u", &uid);
	ret = orasql_execute(obox->ostorage, sth);
	if (ret < 0)
		return FALSE;

	/* update the uids array */
	new_uids = buffer_create_dynamic(system_pool, sizeof(uid) * messages,
					 (size_t)-1);
	while (ret > 0) {
		buffer_append(new_uids, &uid, sizeof(uid));
		ret = orasql_fetch(obox->ostorage, sth);
	}

	if (orasql_close(obox->ostorage, sth) < 0)
		ret = -1;

	if (ret < 0) {
		buffer_free(new_uids);
		return FALSE;
	}

	/* check the differences */
	old_uid = buffer_get_data(obox->uids, &size);
	old_last_uid = old_uid + (size / sizeof(uid));

	new_uid = buffer_get_data(new_uids, &size);
	new_last_uid = new_uid + (size / sizeof(uid));

	for (seq = 1; old_uid != old_last_uid; old_uid++, seq++) {
		while (new_uid == new_last_uid || *old_uid < *new_uid) {
			sc->expunge(&obox->box, seq, sc_context);
			obox->synced_messages--;
			if (new_uid != new_last_uid)
				new_uid++;
		}
	}

	buffer_free(obox->uids);
	obox->uids = new_uids;
	return TRUE;
}

static int oracle_storage_sync_mark_expunges(struct oracle_mailbox *obox,
					     sqlo_stmt_handle_t sth)
{
	unsigned int uid, *old_uid, *old_last_uid;
	size_t size;
	int ret;

	sth = orasql_define_output(obox->ostorage, sth, "%u", &uid);
	ret = orasql_execute(obox->ostorage, sth);
	if (ret < 0)
		return FALSE;

	/* mark the messages expunged by setting their UID to 0 */
	old_uid = buffer_get_modifyable_data(obox->uids, &size);
	old_last_uid = old_uid + (size / sizeof(uid));

	while (ret > 0) {
		while (old_uid != old_last_uid && *old_uid < uid) {
			*old_uid = 0;
			old_uid++;
		}

		if (old_uid == old_last_uid) {
			/* new message */
			buffer_append(obox->uids, &uid, sizeof(uid));
		}
		ret = orasql_fetch(obox->ostorage, sth);
	}

	if (orasql_close(obox->ostorage, sth) < 0 || ret < 0)
		return FALSE;

	return TRUE;
}

static int oracle_storage_sync_expunges(struct oracle_mailbox *obox,
					unsigned int messages,
					int sync_expunges)
{
	sqlo_stmt_handle_t sth;

	/* fetch all UIDs */
	sth = orasql_prepare(obox->ostorage, "SELECT imap_uid FROM mails "
			     "WHERE mailbox_id = %u:id ORDER BY imap_uid",
			     obox->id);

	if (sync_expunges) {
		if (!oracle_storage_sync_and_expunge(obox, sth, messages))
			return FALSE;
	} else {
		if (!oracle_storage_sync_mark_expunges(obox, sth))
			return FALSE;
	}

	return TRUE;
}

static int oracle_storage_sync_new(struct oracle_mailbox *obox,
				   unsigned int messages_count)
{
	sqlo_stmt_handle_t sth;
	unsigned int uid;
	int ret;

	/* fetch new UIDs */
	sth = orasql_prepare(obox->ostorage, "SELECT imap_uid FROM mails "
			     "WHERE mailbox_id = %u:id AND "
			     "imap_uid >= %u:imapuid ORDER BY imap_uid",
			     obox->id, obox->next_uid);

	sth = orasql_define_output(obox->ostorage, sth, "%u", &uid);
	ret = orasql_execute(obox->ostorage, sth);
	if (ret < 0)
		return FALSE;

	if (obox->uids == NULL) {
		obox->uids = buffer_create_dynamic(system_pool,
						   messages_count * 4,
						   (size_t)-1);
	}

	while (ret > 0) {
		buffer_append(obox->uids, &uid, sizeof(uid));
		ret = orasql_fetch(obox->ostorage, sth);
	}

	if (orasql_close(obox->ostorage, sth) < 0 || ret < 0)
		return FALSE;

	return TRUE;
}

static int oracle_storage_sync_flags(struct oracle_mailbox *obox)
{
#if 0
	struct mail_storage_callbacks *sc;
	void *sc_context;
	sqlo_stmt_handle_t sth;
	int ret;

	sc = obox->storage->callbacks;
	sc_context = obox->storage->callback_context;

	/* modseq updated, check flag changes */
	sth = orasql_prepare(obox->ostorage,
			     "SELECT imap_uid, flags FROM mails "
			     "WHERE mailbox_id = %u:id AND "
			     "imap_modseq >= %u:modseq AND "
			     "imap_uid < %u:nextuid",
			     obox->id, obox->next_modseq, obox->next_uid);
	sth = orasql_define_output(obox->ostorage, sth, "%u%u", &uid, &flags);

	ret = orasql_execute(storage, sth);
	while (ret > 0) {
		sc->update_flags(&ibox->box, oracle_mailbox_uid2seq(obox, uid),
				 uid, flags, custom_flags,
				 MAIL_CUSTOM_FLAGS_COUNT, sc_context);
		ret = orasql_fetch(storage, sth);
	}

	if (orasql_close(obox->ostorage, sth) < 0 || ret < 0)
		return FALSE;
#endif
	return TRUE;
}

int oracle_storage_sync(struct mailbox *box, int sync_expunges)
{
	struct oracle_mailbox *obox = (struct oracle_mailbox *)box;
	struct mail_storage *storage = box->storage;
	sqlo_stmt_handle_t sth;
	unsigned int next_uid, next_modseq, next_expungeseq, messages, uid;

	/* get new counters */
	sth = orasql_prepare(obox->ostorage,
			     "SELECT next_uid, next_modseq, "
			     "next_expungeseq, messages_count "
			     "FROM mailboxes WHERE id = %u:id", obox->id);
	sth = orasql_define_output(obox->ostorage, sth, "%u%u%u%u",
				   &next_uid, &next_modseq,
				   &next_expungeseq, &messages);
	if (!orasql_mailbox_fetch_one(obox, sth))
		return FALSE;

	if (next_expungeseq != obox->next_expungeseq &&
	    obox->next_expungeseq != 0) {
		/* messages expunged, fetch UIDs again */
		if (!oracle_storage_sync_expunges(obox, messages,
						  sync_expunges))
			return FALSE;
	} else if (next_uid != obox->next_uid) {
		/* new messages, fetch them */
		i_assert(obox->next_uid < next_uid);
		i_assert(obox->synced_messages <= messages);

		if (!oracle_storage_sync_new(obox, messages))
			return FALSE;
	}

	if (next_modseq != obox->next_modseq && obox->next_modseq != 0) {
		/* flag changes */
		if (!oracle_storage_sync_flags(obox))
			return FALSE;
	}
	obox->next_uid = next_uid;
	obox->next_modseq = next_modseq;
	obox->next_expungeseq = next_expungeseq;

	/* see if there's new messages */
	messages = buffer_get_used_size(obox->uids) / sizeof(uid);
	if (obox->synced_messages != messages) {
		// FIXME: recent
		obox->storage->callbacks->
			new_messages(&obox->box, messages, 0,
				     obox->storage->callback_context);
                obox->synced_messages = messages;
	}

	return TRUE;
}

void oracle_storage_auto_sync(struct mailbox *box,
			      enum mailbox_sync_type sync_type,
			      unsigned int min_newmail_notify_interval)
{
}

