-- users
create table users (
  id integer not null,
  name varchar2(256) not null,
  primary key (id)
);
create unique index user_name_idx on users (name);

create sequence user_id_seq;
create or replace trigger user_id_seq_trigger
before insert on users for each row
begin
  if (:new.id is null) then
    select user_id_seq.nextval into :new.id from dual;
  end if;
end;
/

-- mailboxes
create table mailboxes (
  id integer not null,
  user_id integer not null,
--
  name varchar2(256) not null,
  noselect char(1) default 'N' not null,
--
  uid_validity integer not null,
  next_uid integer default 1 not null,
  next_modseq integer default 1 not null,
  next_expungeseq integer default 1 not null,
--
  messages_count integer default 0 not null,
  seen_messages_count integer default 0 not null,
--  deleted_messages_count integer default 0 not null,
  last_nonrecent_uid integer default 0 not null,
--
--  first_unseen_uid_lowwater integer default 0 not null,
--  first_deleted_uid_lowwater integer default 0 not null,
--
  parent_id integer,
--
  primary key (id),
  foreign key (user_id) references users(id),
  foreign key (parent_id) references mailboxes(id)
);
create unique index user_mailbox_idx on mailboxes (user_id, parent_id, name);

create sequence mailbox_id_seq;
create or replace trigger mailbox_id_seq_trigger
before insert on mailboxes for each row
begin
  if (:new.id is null) then
    select mailbox_id_seq.nextval into :new.id from dual;
  end if;
end;
/

-- custom flags in mailbox
create table mailbox_flags (
  id integer not null,
  mailbox_id integer not null,
  name varchar2(80) not null,
--
  primary key (id),
  foreign key (mailbox_id) references mailboxes(id) on delete cascade
);
create unique index mailbox_flag_idx on mailbox_flags (mailbox_id, name);

create sequence mailbox_flags_id_seq;
create or replace trigger mailbox_flags_id_seq_trigger
before insert on mailbox_flags for each row
begin
  if (:new.id is null) then
    select mailbox_flags_id_seq.nextval into :new.id from dual;
  end if;
end;
/

-- mail body
create table mail_body (
  id integer not null,
  refcount integer default 1 not null,
  data blob,
--
  primary key (id)
) tablespace dovecot_mail_body;

create sequence mail_body_id_seq;
create or replace trigger mail_body_insert_trigger
before insert on mail_body for each row
begin
  if (:new.id is null) then
    select mail_body_id_seq.nextval into :new.id from dual;
  end if;
end;
/

-- mail cache
create table mail_cache (
  mail_body_id integer not null,
--
  sent_date date,
  messageparts long raw,
--
  imap_body long raw,
  imap_bodystructure long raw,
  imap_envelope long raw,
--
  primary key (mail_body_id),
  foreign key (mail_body_id) references mail_body(id) on delete cascade
) tablespace dovecot_cache;

-- mails
create table mails (
  id integer not null,
  mailbox_id integer not null,
  mail_body_id integer not null,
  imap_uid integer not null,
  imap_modseq integer default 0 not null,
--
  internal_date date not null,
  header_size integer not null,
  message_size integer not null,
--
  answered char(1) default 'N' not null,
  flagged char(1) default 'N' not null,
  deleted char(1) default 'N' not null,
  seen char(1) default 'N' not null,
  draft char(1) default 'N' not null,
--
  primary key (id),
  foreign key (mailbox_id) references mailboxes(id),
  foreign key (mail_body_id) references mail_body(id)
) tablespace dovecot_mail;
create unique index mail_uid_idx on mails (mailbox_id, imap_uid);
create index mail_modseq_idx on mails (mailbox_id, imap_modseq);

create sequence mail_id_seq;
create or replace trigger mail_insert_trigger
before insert on mails for each row
declare row rowid;
begin
  if (:new.id is null) then
    select mail_id_seq.nextval into :new.id from dual;
  end if;

  if (:new.imap_uid is null) then
    select rowid, next_uid into row, :new.imap_uid from mailboxes
      where id = :new.mailbox_id for update of next_uid;
    update mailboxes set next_uid = next_uid+1 where rowid = row;
  end if;

  update mailboxes set messages_count = messages_count+1
    where id = :new.mailbox_id;
  if (:new.seen = 'Y') then
    update mailboxes set seen_messages_count = seen_messages_count+1
      where id = :new.mailbox_id;
  end if;
end;
/
create or replace trigger mail_update_trigger
before update on mails for each row
begin
  if (:old.seen = 'N' and :new.seen = 'Y') then
    update mailboxes set seen_messages_count = seen_messages_count+1
      where id = :new.mailbox_id;
  elsif (:old.seen = 'Y' and :new.seen = 'N') then
    update mailboxes set seen_messages_count = seen_messages_count-1
      where id = :new.mailbox_id;
  end if;
end;
/
create or replace trigger mail_delete_trigger
before delete on mails for each row
begin
  if (:old.seen = 'Y') then
    update mailboxes set seen_messages_count = seen_messages_count-1
      where id = :new.mailbox_id;
  end if;
  update mailboxes set messages_count = messages_count-1
    where id = :new.mailbox_id;
end;
/

-- mail custom flags
create table mail_flags (
  mail_id integer not null,
  flag_id integer not null,
--
  primary key (mail_id),
  foreign key (mail_id) references mails(id) on delete cascade,
  foreign key (flag_id) references mailbox_flags(id) on delete cascade
);

-- ...
insert into mailboxes (name, user_id, uid_validity) values ('INBOX', 1, 1);
